# -*- coding: utf-8 -*-
from sqlalchemy import Table, Column, Integer, Unicode, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import column_property, relation, mapper
from sqlalchemy.sql import select


BASE = declarative_base()


class User(BASE):
    __tablename__ = 'user_table'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(50))

    def login(self):
        pass

    def __repr__(self):
        pass


class Admin(User):
    __mapper_args__ = {'polymorphic_identity': 'admin'}

    phone = Column(Unicode(50))

    def permissions(self):
        pass

    def __unicode__(self):
        pass


class Manager(User):
    __mapper_args__ = {'polymorphic_identity': 'manager'}

    department = Column(Unicode(50))

    def permissions(self):
        pass

    def __unicode__(self):
        pass


class Employee(User):
    __mapper_args__ = {'polymorphic_identity': 'employee'}

    manager_id = Column(Integer, ForeignKey(Manager.id))

    # real column
    rank = column_property(Column(Unicode(50)), active_history=True)

    # sql expression label
    department = column_property(
        select([Manager.department]).where(Manager.id == manager_id)
    )


class Address(BASE):
    __tablename__ = 'address_table'

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('user_table.id'))
    user = relation(User, backref="address")


books = Table(
    'books',
    BASE.metadata,
    Column('id', Integer, primary_key=True),
    Column('title', Unicode(50), nullable=False),
    Column('user_id', Integer, ForeignKey('user_table.id')),
)


class Book(object):
    pass


mapper(Book, books, {'user': relation(User, backref='books')})


# Not mapped table
notes = Table(
    'notes',
    BASE.metadata,
    Column('id', Integer, primary_key=True),
    Column('name', Unicode(50), nullable=False),
    Column('user_id', Integer, ForeignKey('user_table.id')),
)



# -*- coding: utf-8 -*-
"""
Program for generating plantuml or dot format
of a database tables by connection string

\n\nDatabase connection string - http://goo.gl/3GpnE
"""
import operator
import string
from optparse import OptionParser
from sqlalchemy import create_engine, MetaData
from sadisplay import describe, render, __version__


def run():
    """Command for reflection database objects"""
    parser = OptionParser(
        version=__version__, description=__doc__,
    )

    parser.add_option(
        '-u', '--url', dest='url',
        help='Database URL (connection string)',
    )

    parser.add_option(
        '-r', '--render', dest='render', default='dot',
        choices=['plantuml', 'dot'],
        help='Output format - plantuml or dot',
    )

    parser.add_option(
        '-l', '--list', dest='list', action='store_true',
        help='Output database list of tables and exit',
    )

    parser.add_option(
        '-i', '--include', dest='include',
        help='List of tables to include through ","',
    )

    parser.add_option(
        '-e', '--exclude', dest='exclude',
        help='List of tables to exlude through ","',
    )

    (options, args) = parser.parse_args()

    if not options.url:
        print('-u/--url option required')
        exit(1)

    engine = create_engine(options.url)
    meta = MetaData()

    meta.reflect(bind=engine)

    if options.list:
        print('Database tables:')
        tables = sorted(meta.tables.keys())

        def _g(l, i):
            try:
                return tables[i]
            except IndexError:
                return ''

        for i in range(0, len(tables), 2):
            print(' {0}{1}{2}'.format(
                _g(tables, i),
                ' ' * (38 - len(_g(tables, i))),
                _g(tables, i + 1),
            ))

        exit(0)

    tables = set(meta.tables.keys())

    if options.include:
        tables &= set(map(string.strip, options.include.split(',')))

    if options.exclude:
        tables -= set(map(string.strip, options.exclude.split(',')))

    desc = describe(map(lambda x: operator.getitem(meta.tables, x), tables))
    print(getattr(render, options.render)(desc))



# -*- coding: utf-8 -*-
from nose.tools import assert_equal
import sadisplay
import model


class TestDescribe(object):

    def test_single_mapper(self):

        objects, relations, inherits = sadisplay.describe([model.User])

        assert len(objects) == 1
        assert relations == []
        assert inherits == []
        assert_equal(
            objects[0],
            {
                'name': model.User.__name__,
                'cols': [
                    ('INTEGER', 'id', 'pk'),
                    ('VARCHAR(50)', 'name', None),
                ],
                'props': ['address', 'books', ],
                'methods': ['login', ],
            }
        )

    def test_single_table(self):

        objects, relations, inherits = sadisplay.describe([model.notes])

        assert len(objects) == 1
        assert relations == []
        assert inherits == []
        assert_equal(
            objects[0],
            {
                'name': model.notes.name,
                'cols': [
                    ('INTEGER', 'id', 'pk'),
                    ('VARCHAR(50)', 'name', None),
                    ('INTEGER', 'user_id', 'fk')
                ],
                'props': [],
                'methods': [],
            }
        )

    def test_inherits(self):

        objects, relations, inherits = sadisplay \
            .describe([model.User, model.Admin, model.Manager])

        assert len(relations) == 0
        assert len(objects) == 3
        assert len(inherits) == 2
        assert_equal(
            objects[1],
            {
                'name': model.Admin.__name__,
                'cols': [
                    ('INTEGER', 'id', 'pk'),
                    ('VARCHAR(50)', 'name', None),
                    ('VARCHAR(50)', 'phone', None),
                ],
                'props': ['address', 'books', ],
                'methods': ['permissions', ],
            }
        )

        assert_equal(
            inherits[0],
            {
                'child': model.Admin.__name__,
                'parent': model.User.__name__,
            }
        )

    def test_relation(self):

        objects, relations, inherits = sadisplay \
            .describe([model.User, model.Address])

        assert len(objects) == 2
        assert_equal(
            objects[1],
            {
                'name': model.Address.__name__,
                'cols': [
                    ('INTEGER', 'id', 'pk'),
                    ('INTEGER', 'user_id', 'fk'),
                ],
                'props': ['user'],
                'methods': [],
            }
        )

        assert len(inherits) == 0
        assert_equal(
            relations[0],
            {
                'from': model.Address.__name__,
                'to': model.User.__name__,
                'by': 'user_id',
            }
        )

    def test_table(self):

        objects, relations, inherits = sadisplay \
            .describe([model.Book])

        assert len(objects) == 1
        assert_equal(
            objects[0],
            {
                'name': model.Book.__name__,
                'cols': [
                    ('INTEGER', 'id', 'pk'),
                    ('VARCHAR(50)', 'title', None),
                    ('INTEGER', 'user_id', 'fk'),
                ],
                'props': ['user'],
                'methods': [],
            }
        )

        objects, relations, inherits = sadisplay \
            .describe([model.Book, model.books])

        assert len(objects) == 1
        assert_equal(
            objects[0],
            {
                'name': model.Book.__name__,
                'cols': [
                    ('INTEGER', 'id', 'pk'),
                    ('VARCHAR(50)', 'title', None),
                    ('INTEGER', 'user_id', 'fk'),
                ],
                'props': ['user'],
                'methods': [],
            }
        )

        objects, relations, inherits = sadisplay \
            .describe([model.books])

        assert len(objects) == 1
        assert_equal(
            objects[0],
            {
                'name': model.books.name,
                'cols': [
                    ('INTEGER', 'id', 'pk'),
                    ('VARCHAR(50)', 'title', None),
                    ('INTEGER', 'user_id', 'fk'),
                ],
                'props': [],
                'methods': [],
            }
        )

    def test_column_property(self):

        objects, relations, inherits = sadisplay \
            .describe([model.Employee])

        assert_equal(len(objects), 1)
        assert_equal(
            objects[0],
            {
                'name': model.Employee.__name__,
                'cols': [
                    ('INTEGER', 'id', 'pk'),
                    ('INTEGER', 'manager_id', 'fk'),
                    ('VARCHAR(50)', 'name', None),
                    ('VARCHAR(50)', 'rank', None),
                ],
                'props': ['address', 'books', 'department'],
                'methods': [],
            })



# -*- coding: utf-8 -*-
import uuid
import types
import locale
from functools import cmp_to_key

from sqlalchemy import exc, orm
from sqlalchemy.orm import class_mapper
from sqlalchemy import Column, Integer, Table
from sqlalchemy.orm.properties import ColumnProperty

try:
    # sa >= 0.9
    from sqlalchemy.sql.elements import Label
except ImportError:
    from sqlalchemy.sql.expression import Label


def describe(items, show_methods=True, show_properties=True):
    """Detecting attributes, inherits and relations

    :param items: list of objects to describe
    :param show_methods: do detection of methods
    :param show_properties: do detection of properties

    Return tuple (objects, relations, inherits)


    Where objects is list::

        [{
            'name': '<Mapper class name or table name>',
            'cols': [
                (
                    '<Column type class name>',
                    '<Column name>',
                    '<pk|fk>',),
                ...
            ],
            'props': ['<Property name>'],
            'methods': ['<Method name>', ...],
        }, ...]


    Relations is::

        [{
            'from': '<From mapper class name>',
            'by': '<By mapper foreign key column name>',
            'to': '<To mapper class name>',
        }, ...]


    Example usage::

        import sadisplay
        from app import models

        desc = sadisplay.describe([
            getattr(model, attr) for attr in dir(model)
        ])

        desc = sadisplay.describe([models.User, models.Group])
    """

    def column_role(column):
        if column.primary_key:
            return 'pk'
        elif column.foreign_keys:
            return 'fk'

    def column_compare(a, b):
        if a[2] != 'pk' and b[2] == 'pk':
            return 1
        if a[2] == 'pk' and b[2] != 'pk':
            return -1
        return locale.strcoll(a[1], b[1])

    class EntryItem(object):
        """Class adaptor for mapped classes and tables"""
        name = None
        methods = []
        columns = []
        inherits = None
        properties = []
        bases = tuple()

        def __init__(self, mapper=None, table=None):

            if mapper is not None:
                self.name = mapper.class_.__name__
                self.columns = mapper.columns
                self.methods = mapper.class_.__dict__.items()
                self.inherits = mapper.inherits
                self.properties = mapper.iterate_properties
                self.bases = mapper.class_.__bases__
                self.class_ = mapper.class_
                self.table_name = str(mapper.mapped_table)

            elif table is not None:
                self.name = table.name
                self.table_name = table.name
                # prepend schema if exists for foreign key matching
                if hasattr(table, "schema") and table.schema:
                    self.table_name = table.schema + "." + self.table_name
                self.columns = table.columns
            else:
                pass

        def __repr__(self):
            return '<{s.__class__.__name__} {s.name}>'.format(s=self)

        def __eq__(self, other):
            if other.inherits or self.inherits:
                return self.name == other.name
            return self.table_name == other.table_name

    objects = []
    relations = []
    inherits = []

    entries = []
    for item in items:
        try:
            mapper = class_mapper(item)
        except (exc.ArgumentError, orm.exc.UnmappedClassError):
            if isinstance(item, Table):
                entity = EntryItem(table=item)
            else:
                continue
        else:
            entity = EntryItem(mapper=mapper)

        if entity not in entries:
            entries.append(entity)

    for entry in entries:

        result_item = {
            'name': entry.name,
            'cols': [
                (
                    str(col.type),
                    name,
                    column_role(col),
                ) for name, col in entry.columns.items()
                if not isinstance(col, Label)
            ],
            'props': [],
            'methods': [],
        }

        # sort columns by role and name
        result_item['cols'].sort(
            key=cmp_to_key(column_compare),
        )

        if show_methods and entry.methods:

            if entry.inherits:
                base_methods = entry.inherits.class_.__dict__.keys()
            else:
                # Create the DummyClass subclass of mapper bases
                # for detecting mapper own methods
                suffix = '%s' % str(uuid.uuid4())
                params = {
                    '__tablename__': 'dummy_table_%s' % suffix,
                    'dummy_id_col': Column(Integer, primary_key=True)
                }

                DummyClass = type('Dummy%s' % suffix, entry.bases, params)

                base_methods = DummyClass.__dict__.keys()

            # Filter mapper methods
            for name, func in entry.methods:
                if name[0] != '_' and name not in base_methods:
                    if isinstance(func, types.FunctionType):
                        result_item['methods'].append(name)

        if show_properties:

            # find relationship properties
            for item in entry.properties:
                if not isinstance(item, ColumnProperty):
                    result_item['props'].append(item.key)

            # find column_property() as sql expressions
            for name, item in entry.columns.items():
                if isinstance(item, Label):
                    result_item['props'].append(name)

        # ordering
        for key in ('methods', 'props'):
            result_item[key].sort()

        objects.append(result_item)

        # Detect relations by ForeignKey
        for col in entry.columns:
            for fk in col.foreign_keys:
                table = fk.column.table
                for m in entries:
                    try:
                        if str(table) == str(m.table_name):
                            relations.append({
                                'from': entry.name,
                                'by': col.name,
                                'to': m.name,
                            })
                    except AttributeError:
                        pass

        if entry.inherits:

            inh = {
                'child': entry.name,
                'parent': EntryItem(mapper=entry.inherits).name,
            }

            inherits.append(inh)

            # Delete relation by inherits
            for i, rel in enumerate(relations):
                if inh['child'] == rel['from'] and inh['parent'] == rel['to']:
                    relations.pop(i)

    return objects, relations, inherits



# -*- coding: utf-8 -*-
import re
import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


here = os.path.dirname(os.path.abspath(__file__))
version_re = re.compile(r"__version__ = (\'.*?\')")
with open(os.path.join(os.path.join(here, 'sadisplay'), '__init__.py')) as f:
    version = None
    for line in f:
        match = version_re.search(line)
        if match:
            version = eval(match.group(1))
            break
    else:
        raise Exception("Cannot find version in sadisplay.py")


setup(
    name='sadisplay',
    version=version,
    url='http://bitbucket.org/estin/sadisplay',
    license='BSD',
    author='Evgeniy Tatarkin',
    author_email='tatarkin.evg@gmail.com',
    description='SqlAlchemy schema display script',
    long_description=open(os.path.join(here, 'README.rst')).read(),
    packages=['sadisplay', ],
    zip_safe=False,
    platforms='any',
    tests_require=['nose >= 1.0', ],
    test_suite='nose.collector',
    install_requires=[
        'SQLAlchemy >= 0.5',
    ],
    entry_points={
        'console_scripts': ['sadisplay = sadisplay.reflect:run', ],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
)



# -*- coding: utf-8 -*-
from sadisplay import __version__


# by pull request
# https://bitbucket.org/estin/sadisplay/pull-requests/4/format-table-info/diff
def format_column(column):
    type, name, role = column
    role_char = {
        'pk': u'\U00002605',
        'fk': u'\U00002606',
    }.get(role, u'\U000026AA')

    return type, '%s %s' % (role_char, name)


def format_property(property_name):
    type_char = u'\U000026AA'

    return '%s %s' % (type_char, property_name)


def tabular_output(table, indent=None, col_delimiter=None):
    indent = indent or ' ' * 4
    col_delimiter = col_delimiter or ' '
    col_width = [max(len(x) for x in col) for col in zip(*table)]
    for line in table:
        yield indent + col_delimiter.join(
            '{:{}}'.format(x, col_width[i]) for i, x in enumerate(line)
        )


def plantuml(desc):
    """Generate plantuml class diagram

    :param desc: result of sadisplay.describe function

    Return plantuml class diagram string
    """

    classes, relations, inherits = desc

    result = [
        '@startuml',
        'skinparam defaultFontName Courier',
    ]

    def _cleanup(col):
        type, name = col
        return type.replace('(', '[').replace(')', ']'), name

    for cls in classes:
        # issue #11 - tabular output of class members (attrs)
        # http://stackoverflow.com/a/8356620/258194

        # build table
        class_desc = []
        # table columns
        class_desc += [
            _cleanup(format_column(i)) for i in cls['cols']
        ]
        # class properties
        class_desc += [('+', i) for i in cls['props']]
        # methods
        class_desc += [('%s()' % i, '') for i in cls['methods']]

        result.append(
            'Class %(name)s {\n%(desc)s\n}' % {
                'name': cls['name'],
                'desc': '\n'.join(tabular_output(class_desc)),
            }
        )

    for item in inherits:
        result.append("%(parent)s <|-- %(child)s" % item)

    for item in relations:
        result.append("%(from)s <--o %(to)s: %(by)s" % item)

    result += [
        'right footer generated by sadisplay v%s' % __version__,
        '@enduml',
    ]

    return '\n\n'.join(result)


def dot(desc):
    """Generate dot file

    :param desc: result of sadisplay.describe function

    Return string
    """

    classes, relations, inherits = desc

    CLASS_TEMPLATE = """
        %(name)s [label=<
        <TABLE BGCOLOR="lightyellow" BORDER="0"
            CELLBORDER="0" CELLSPACING="0">
                <TR><TD COLSPAN="2" CELLPADDING="4"
                        ALIGN="CENTER" BGCOLOR="palegoldenrod"
                ><FONT FACE="Helvetica Bold" COLOR="black"
                >%(name)s</FONT></TD></TR>%(cols)s%(props)s%(methods)s
        </TABLE>
    >]
    """

    COLUMN_TEMPLATE = """<TR><TD ALIGN="LEFT" BORDER="0"
        ><FONT FACE="Bitstream Vera Sans">%(name)s</FONT
        ></TD><TD ALIGN="LEFT"
        ><FONT FACE="Bitstream Vera Sans">%(type)s</FONT
        ></TD></TR>"""

    PROPERTY_TEMPLATE = """<TR><TD ALIGN="LEFT" BORDER="0"
        BGCOLOR="palegoldenrod"
        ><FONT FACE="Bitstream Vera Sans">%(name)s</FONT></TD
        ><TD BGCOLOR="palegoldenrod" ALIGN="LEFT"
        ><FONT FACE="Bitstream Vera Sans">PROPERTY</FONT
        ></TD></TR>"""

    METHOD_TEMPLATE = """<TR><TD ALIGN="LEFT" BORDER="0"
        BGCOLOR="palegoldenrod"
        ><FONT FACE="Bitstream Vera Sans">%(name)s()</FONT></TD
        ><TD BGCOLOR="palegoldenrod" ALIGN="LEFT"
        ><FONT FACE="Bitstream Vera Sans">METHOD</FONT
        ></TD></TR>"""

    EDGE_INHERIT = "\tedge [\n\t\tarrowhead = empty\n\t]"
    INHERIT_TEMPLATE = "\t%(child)s -> %(parent)s \n"

    EDGE_REL = "\tedge [\n\t\tarrowhead = ediamond\n\t\tarrowtail = open\n\t]"
    RELATION_TEMPLATE = "\t\"%(from)s\" -> \"%(to)s\" [label = \"%(by)s\"]"

    result = ["""
        digraph G {
            label = "generated by sadisplay v%s";
            fontname = "Bitstream Vera Sans"
            fontsize = 8

            node [
                fontname = "Bitstream Vera Sans"
                fontsize = 8
                shape = "plaintext"
            ]

            edge [
                fontname = "Bitstream Vera Sans"
                fontsize = 8
            ]
    """ % __version__]

    for cls in classes:
        cols = ' '.join([
            COLUMN_TEMPLATE % {'type': c[0], 'name': c[1]}
            for c in map(format_column, cls['cols'])
        ])
        props = ' '.join([
            PROPERTY_TEMPLATE % {'name': format_property(p)}
            for p in cls['props']
        ])
        methods = ' '.join([
            METHOD_TEMPLATE % {'name': m} for m in cls['methods']
        ])
        renderd = CLASS_TEMPLATE % {
            'name': cls['name'],
            'cols': cols,
            'props': props,
            'methods': methods,
        }

        result.append(renderd)

    result += [EDGE_INHERIT]
    for item in inherits:
        result.append(INHERIT_TEMPLATE % item)

    result += [EDGE_REL]
    for item in relations:
        result.append(RELATION_TEMPLATE % item)

    result += [
        '}'
    ]

    return '\n'.join(result)
